// up2k.go - resumable multithreaded http uploads
// ed <irc.rizon.net>, MIT-licensed
// https://ocv.me/dev/?up2k.go
// 2019-05-15, 21:30, v0.1
//
// usage example:
//   curl -F "f=@Various.zip" http://127.0.0.1:1234/
//
// remove hasher from the multiwriter for a ~2.7x speed boost
// however that also gives the python ver a ~4.1x boost
// ( comment hashobj.update(buf) and uncomment /dev/zero
//    in https://ocv.me/dev/?filedrop-1.0.py )
//
//  go with sha1:  444 MiB/s
// py2 with sha1:  555 MiB/s
//  go sans sha1: 1227 MiB/s
// py2 sans sha1: 2309 MiB/s

package main

import (
	"crypto/sha1"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"os"
	"time"
)

func chk(e error) {
	if e != nil {
		panic(e)
	}
}

func rlog(r *http.Request, msg string) {
	//time.Now().Format("2006-01-02 15:04:05"),
	log.Printf("%s %s\n", r.RemoteAddr, msg)
}

func rflog(r *http.Request, format string, args ...interface{}) {
	//time.Now().Format("2006-01-02 15:04:05"),
	log.Printf("%s "+format, r.RemoteAddr, args)
}

func handleRequest(w http.ResponseWriter, r *http.Request) {
	rlog(r, r.Method+" "+r.RequestURI)
	t0 := time.Now().UnixNano()

	// port these APIs:
	//   xhr.open('POST', 'handshake.php', true);
	//   xhr.open('POST', 'chunkpit.php', true);
	//   and index.php

	if r.Method == "GET" && r.RequestURI == "/" {
		buf, err := ioutil.ReadFile("index.html")
		chk(err)
		w.Write(buf)

	} else if r.Method == "POST" {
		r.Body = http.MaxBytesReader(w, r.Body, 1<<32)
		reader, err := r.MultipartReader()
		chk(err)

		hasher := sha1.New()
		part, err := reader.NextPart()
		rflog(r, "receiving %s\n", part.FileName())

		outfile, err := os.Create("/dev/null")
		defer outfile.Close()

		hashwriter := io.MultiWriter(outfile, hasher)
		written, err := io.Copy(hashwriter, part)
		/*
			buf := make([]byte, 1024*64)
			written := 0
			for true {
				nread, err := part.Read(buf)
				if nread <= 0 || err != nil {
					break
				}
				written += nread
				nw1, err1 := hasher.Write(buf[:nread])
				nw2, err2 := outfile.Write(buf[:nread])
				if nread != nw1 || nread != nw2 || err1 != nil || err2 != nil {
					panic("truncated write to hasher or outfile")
				}
			}
		*/
		hash := hasher.Sum(nil)

		t := time.Now().UnixNano()
		td := float64(t-t0) / (1000 * 1000 * 1000)
		spd := float64(written) / (td * 1024 * 1024)

		msg := fmt.Sprintf("got %d bytes in %f seconds, %f MiB/s, %x",
			written, td, spd, hash)

		rlog(r, msg)
		fmt.Fprintf(w, msg+"\n")

	} else if r.Method == "POST_MISTAKE_1" {
		infile, header, err := r.FormFile("f")
		chk(err)
		defer infile.Close()

		fmt.Printf("receiving %s\n", header.Filename)
		outfile, err := os.Create("/dev/null")
		chk(err)

		io.Copy(outfile, infile)
		fmt.Fprintf(w, "ok")

	} else {
		w.WriteHeader(404)
		fmt.Fprintf(w, "not found")
	}
}

func main() {
	bindIP := flag.String("ip", "0.0.0.0", "bind to ip")
	bindPort := flag.Int("port", 1234, "bind to port")
	//nullwrite = flag.Bool("null", false, "disable file writes")
	flag.Parse()

	localIP := "127.0.0.1"
	conn, err := net.Dial("udp", "10.255.255.255:1")
	if *bindIP != "127.0.0.1" && err == nil {
		localIP = conn.LocalAddr().(*net.UDPAddr).IP.String()
	}

	localEp := fmt.Sprintf("%s:%d", localIP, *bindPort)
	bindEp := fmt.Sprintf("%s:%d", *bindIP, *bindPort)

	http.HandleFunc("/", handleRequest)

	fmt.Println("listening on " + bindEp)
	fmt.Println("available at http://" + localEp)
	fmt.Println(http.ListenAndServe(bindEp, nil))
}
